import numpy as np
import matplotlib.pyplot as plt
import matplotlib

matplotlib.rc('font',size=14)

#read input
d = np.genfromtxt('../sims/test.pdr.fin')
Av = d[:,2]; Tgas = d[:,3]
HI = d[:,39]; H2 = d[:,38]
CII = d[:,18]; CI = d[:,32]; CO = d[:,35]

#gas temperature profile
plt.figure(figsize=(8,6))
plt.grid(True);plt.grid(True,which='minor',lw=0.6,ls='--')
plt.semilogx(Av,Tgas,'-',lw=2)
plt.xlabel(r'A$_{\rm V}$ [mag]')
plt.ylabel('Gas temperature [K]')
plt.title('Gas temperature profile',weight='bold')
#plt.savefig('Tgas.png',bbox_inches='tight')

#HI/H2 transition
plt.figure(figsize=(8,6))
plt.grid(True);plt.grid(True,which='minor',lw=0.6,ls='--')
plt.loglog(Av,HI,'-',lw=2,label='HI')
plt.loglog(Av,H2,'-',lw=2,label=r'H$_2$')
plt.legend(loc=0)
plt.xlabel(r'A$_{\rm V}$ [mag]')
plt.ylabel('Relative abundance')
plt.title(r'HI-to-H$_2$ transition',weight='bold')
#plt.savefig('HItoH2.png',bbox_inches='tight')

#carbon cycle
plt.figure(figsize=(8,6))
plt.grid(True);plt.grid(True,which='minor',lw=0.6,ls='--')
plt.loglog(Av,CII,'-',lw=2,label='CII')
plt.loglog(Av,CI,'-',lw=2,label='CI')
plt.loglog(Av,CO,'-',lw=2,label='CO')
plt.legend(loc=0)
plt.xlabel(r'A$_{\rm V}$ [mag]')
plt.ylabel('Relative abundance')
plt.title(r'Carbon cycle',weight='bold')
#plt.savefig('CarbonCycle.png',bbox_inches='tight')

plt.show()
